import torch
import torch.nn as nn


class ConvBlock(nn.Module):
    def __init__(self, in_channels, mid_channels, out_channels):
        super().__init__()
        self.activation = nn.ReLU(inplace=True)
        self.conv1 = nn.Conv2d(in_channels, mid_channels, kernel_size=3, padding=1, bias=True)
        self.bn1 = nn.BatchNorm2d(mid_channels)
        self.conv2 = nn.Conv2d(mid_channels, out_channels, kernel_size=3, padding=1, bias=True)
        self.bn2 = nn.BatchNorm2d(out_channels)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.activation(x)

        x = self.conv2(x)
        x = self.bn2(x)
        output = self.activation(x)

        return output


class UNetPlusPlus(nn.Module):
    def __init__(self, in_channels=1, out_channels=1):
        super(UNetPlusPlus, self).__init__()
        self.out_channels = out_channels

        n1 = 64
        filters = [n1, n1 * 2, n1 * 4, n1 * 8, n1 * 16]

        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
        self.up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)

        self.conv0_0 = ConvBlock(in_channels, filters[0], filters[0])
        self.conv1_0 = ConvBlock(filters[0], filters[1], filters[1])
        self.conv2_0 = ConvBlock(filters[1], filters[2], filters[2])
        self.conv3_0 = ConvBlock(filters[2], filters[3], filters[3])
        self.conv4_0 = ConvBlock(filters[3], filters[4], filters[4])

        self.conv0_1 = ConvBlock(filters[0] + filters[1], filters[0], filters[0])
        self.conv1_1 = ConvBlock(filters[1] + filters[2], filters[1], filters[1])
        self.conv2_1 = ConvBlock(filters[2] + filters[3], filters[2], filters[2])
        self.conv3_1 = ConvBlock(filters[3] + filters[4], filters[3], filters[3])

        self.conv0_2 = ConvBlock(filters[0] * 2 + filters[1], filters[0], filters[0])
        self.conv1_2 = ConvBlock(filters[1] * 2 + filters[2], filters[1], filters[1])
        self.conv2_2 = ConvBlock(filters[2] * 2 + filters[3], filters[2], filters[2])

        self.conv0_3 = ConvBlock(filters[0] * 3 + filters[1], filters[0], filters[0])
        self.conv1_3 = ConvBlock(filters[1] * 3 + filters[2], filters[1], filters[1])

        self.conv0_4 = ConvBlock(filters[0] * 4 + filters[1], filters[0], filters[0])

        self.final = nn.Conv2d(filters[0], out_channels, kernel_size=1)

    def forward(self, x):
        x0_0 = self.conv0_0(x)
        x1_0 = self.conv1_0(self.pool(x0_0))
        x0_1 = self.conv0_1(torch.cat([x0_0, self.up(x1_0)], 1))

        x2_0 = self.conv2_0(self.pool(x1_0))
        x1_1 = self.conv1_1(torch.cat([x1_0, self.up(x2_0)], 1))
        x0_2 = self.conv0_2(torch.cat([x0_0, x0_1, self.up(x1_1)], 1))

        x3_0 = self.conv3_0(self.pool(x2_0))
        x2_1 = self.conv2_1(torch.cat([x2_0, self.up(x3_0)], 1))
        x1_2 = self.conv1_2(torch.cat([x1_0, x1_1, self.up(x2_1)], 1))
        x0_3 = self.conv0_3(torch.cat([x0_0, x0_1, x0_2, self.up(x1_2)], 1))

        x4_0 = self.conv4_0(self.pool(x3_0))
        x3_1 = self.conv3_1(torch.cat([x3_0, self.up(x4_0)], 1))
        x2_2 = self.conv2_2(torch.cat([x2_0, x2_1, self.up(x3_1)], 1))
        x1_3 = self.conv1_3(torch.cat([x1_0, x1_1, x1_2, self.up(x2_2)], 1))
        x0_4 = self.conv0_4(torch.cat([x0_0, x0_1, x0_2, x0_3, self.up(x1_3)], 1))

        output = self.final(x0_4)
        return output



import json
import os.path
from base64 import b64encode, b64decode
from collections import OrderedDict
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd
from numpy.lib.format import dtype_to_descr, descr_to_dtype
from openpyxl import load_workbook

from jbag import logger


def read_mat(input_file, key='scene'):
    from scipy.io import loadmat
    data = loadmat(input_file)[key]
    return data


def save_mat(output_file, data, key='scene'):
    from scipy.io import savemat
    ensure_dir(output_file)
    savemat(output_file, {key: data})


def read_txt_2_list(input_file):
    with open(input_file, 'r') as input_file:
        return [each.strip('\n') for each in input_file.readlines()]


def write_list_2_txt(output_file, data_lst):
    ensure_dir(output_file)
    with open(output_file, 'w') as file:
        for i in range(len(data_lst)):
            file.write(data_lst[i])
            if i != len(data_lst) - 1:
                file.write('\n')


def save_nifti(output_file,
               data,
               voxel_spacing=None,
               orientation='LPI'):
    """
    Save image with nii format.

    Args:
        output_file (str or pathlib.Path):
        data (numpy.ndarray):
        voxel_spacing (sequence or None, optional, default=None): `tuple(x, y, z)`. Voxel spacing of each axis. If None,
            make `voxel_spacing` as `(1.0, 1.0, 1.0)`.
        orientation (str, optional, default="LPI"): "LPI" | "ARI". LPI: Left-Posterior-Inferior;
            ARI: Anterior-Right-Inferior.

    Returns:

    """
    if voxel_spacing is None:
        voxel_spacing = (1.0, 1.0, 1.0)  # replace this with your desired voxel spacing in millimeters

    match orientation:
        case 'LPI':
            affine_matrix = np.diag(list(voxel_spacing) + [1.0])
        case 'ARI':
            # calculate the affine matrix based on the desired voxel spacing and ARI orientation
            affine_matrix = np.array([
                [0, -voxel_spacing[0], 0, 0],
                [-voxel_spacing[1], 0, 0, 0],
                [0, 0, voxel_spacing[2], 0],
                [0, 0, 0, 1]
            ])
        case _:
            raise ValueError(f'Unsupported orientation {orientation}.')

    # create a NIfTI image object
    import nibabel as nib
    ensure_dir(output_file)
    nii_img = nib.Nifti1Image(data, affine=affine_matrix)
    nib.save(nii_img, output_file)


def read_dicom_series(input_dir: Union[str, Path]):
    from pydicom import dcmread
    if not os.path.exists(input_dir):
        raise FileExistsError(f'{input_dir} does not exist.')
    instances = []
    for each in os.listdir(input_dir):
        if each.endswith('.dcm'):
            instances.append(each)

    instances.sort()
    images = []
    for slice_file_name in instances:
        slice_file = os.path.join(input_dir, slice_file_name)
        dicom_data = dcmread(slice_file)
        if 'PixelData' in dicom_data:
            pixel_data = dicom_data.pixel_array
            images.append(pixel_data)

    return np.stack(images)


# JSON
def np_object_hook(dct):
    """
    Convert JSON list or scalar to numpy.

    Args:
        dct (mapping):

    Returns:

    """
    if '__ndarray__' in dct:
        shape = dct['shape']
        dtype = descr_to_dtype(dct['dtype'])
        if shape:
            order = 'C' if dct['Corder'] else 'F'
            if dct['base64']:
                np_obj = np.frombuffer(b64decode(dct['__ndarray__']), dtype=dtype)
                np_obj = np_obj.copy(order=order)
            else:
                np_obj = np.asarray(dct['__ndarray__'], dtype=dtype, order=order)
            return np_obj.reshape(shape)

        if dct['base64']:
            np_obj = np.frombuffer(b64decode(dct['__ndarray__']), dtype=dtype)[0]
        else:
            t = getattr(np, dtype.name)
            np_obj = t(dct['__ndarray__'])
        return np_obj

    return dct


def read_json(input_json_file):
    """
    Read and convert `input_json_file`.

    Args:
        input_json_file (str or pathlib.Path):

    Returns:

    """
    if not os.path.exists(input_json_file):
        raise FileExistsError(f'{input_json_file} does not exist.')
    with open(input_json_file, 'r') as json_file:
        dct = json.load(json_file, object_hook=np_object_hook)
    return dct


class NumpyJSONEncoder(json.JSONEncoder):
    def __init__(self, primitive=False, base64=True, **kwargs):
        """
        JSON encoder for `numpy.ndarray`.

        Args:
            primitive (bool, optional, default=False): Use primitive type if `True`. In primitive schema,
                `numpy.ndarray` is stored as JSON list and `np.generic` is stored as a number.
            base64 (bool, optional, default=True): Use base64 to encode.
            **kwargs:
        """
        self.primitive = primitive
        self.base64 = base64
        super().__init__(**kwargs)

    def default(self, obj):
        if isinstance(obj, (np.ndarray, np.generic)):
            if self.primitive:
                return obj.tolist()
            else:
                if self.base64:
                    data_json = b64encode(obj.data if obj.flags.c_contiguous else obj.tobytes()).decode("ascii")
                else:
                    data_json = obj.tolist()
                dct = OrderedDict(__ndarray__=data_json,
                                  dtype=dtype_to_descr(obj.dtype),
                                  shape=obj.shape,
                                  Corder=obj.flags['C_CONTIGUOUS'],
                                  base64=self.base64)
                return dct
        return super().default(obj)


def save_json(output_file, obj, primitive=False, base64=True):
    """
    Convert obj to JSON object and save as file.

    Args:
        output_file (str or pathlib.Path):
        obj (mapping):
        primitive (bool, optional, default=False): Use primitive type if `True`. In primitive schema, `numpy.ndarray` is
            stored as JSON list and `np.generic` is stored as a number.
        base64 (bool, optional, default=True): Use base64 to encode.

    Returns:

    """
    ensure_dir(output_file)
    with open(output_file, 'w') as file:
        json.dump(obj, file, cls=NumpyJSONEncoder, **{'primitive': primitive, 'base64': base64})


def ensure_dir(input_file):
    input_dir = os.path.split(input_file)[0]
    if not os.path.exists(input_dir):
        os.makedirs(input_dir, exist_ok=True)


def scp(dst_user, dst_host, dst_path, local_path, dst_port=None, recursive=False, send=False, receive=False):
    """
    Transmit file(s) through scp.

    Args:
        dst_user (str):
        dst_host (str):
        dst_path (str or Path):
        local_path (str or Path):
        dst_port (str or int or None, optional, default=None): If None, usually refer to port 22.
        recursive (bool, default=False): Transmit directories recursively.
        send (bool, default=False): Send file(s) from local to destination.
        receive (bool, default=False): Receive file(s) sent from destination.

    Returns:

    """
    assert send ^ receive

    cmd = 'scp'
    dst = f'{dst_user}@{dst_host}:{dst_path}'
    if recursive:
        cmd += ' -r'
    if dst_port:
        cmd += f' -P {dst_port}'
    if send:
        cmd += f' {local_path} {dst}'
    else:
        cmd += f' {dst} {local_path}'
    os.system(cmd)


def save_excel(output_file, data: Union[dict, pd.DataFrame], sheet_name: str = 'Sheet1', append: bool = False, overlay_sheet: bool=False,
               column_width: int = None, auto_adjust_width: bool = False, index=False):
    """
    Save data to Excel file.
    Args:
        output_file (str):
        data (dict | pd.DataFrame):
        sheet_name (str, optional, default='Sheet1'):
        append (bool, optional, default=False): If True, append to existing file.
        overlay_sheet (bool, optional, default=False): If True, overwrite existing sheet. Note that this option only works for appending mode.
        column_width (int, optional, default=None): Set column width to the given value, if exist.
        auto_adjust_width (bool, optional, default=False): If True, adjust column width according to the content length.
        index (bool, optional, default=False): If True, set index column on the work sheet.

    Returns:

    """
    write_mode = 'a' if append else 'w'

    if write_mode == 'a' and not os.path.exists(output_file):
        logger.warning(f'Try to append data to a non-existing file: {output_file}, change mode to write instead.')
        write_mode = 'w'

    if write_mode == 'a' and overlay_sheet:
        if_sheet_exists = 'overlay'
    else:
        if_sheet_exists = None

    if isinstance(data, dict):
        data = pd.DataFrame(data)

    with pd.ExcelWriter(output_file, engine='openpyxl', mode=write_mode, if_sheet_exists=if_sheet_exists) as writer:
        data.to_excel(writer, sheet_name=sheet_name, index=index)

    if column_width is not None or auto_adjust_width:
        book = load_workbook(output_file)
        worksheet = book[sheet_name]

        for column_cells in worksheet.columns:
            if column_width is not None:
                width = column_width
            else:
                max_length = max(len(str(cell.value)) for cell in column_cells)
                width = max_length + 5

            column_letter = column_cells[0].column_letter
            worksheet.column_dimensions[column_letter].width = width

        # Save the workbook
        book.save(output_file)


import torch

from jbag.transforms.transforms import RandomTransform


class MirrorTransform(RandomTransform):
    def __init__(self, keys,
                 apply_probability,
                 allowed_axes,
                 p_per_axes: float = 0.5):
        """
        Mirror transform.
        Args:
            keys (str or sequence):
            apply_probability (float):
            allowed_axes (int or sequence): Axis(es) for mirroring.
            p_per_axes (bool, optional, default=False): Probability for performing transform on each axis.

        """
        super().__init__(keys, apply_probability)
        if isinstance(allowed_axes, int):
            allowed_axes = [allowed_axes]
        self.allowed_axes = allowed_axes
        self.p_per_axes = p_per_axes

    def _call_fun(self, data):
        allowed_axes = [axis for axis in self.allowed_axes if torch.rand(1) < self.p_per_axes]
        if len(allowed_axes) == 0:
            return data
        for key in self.keys:
            value: torch.Tensor = data[key]
            value = torch.flip(value, allowed_axes)
            data[key] = value
        return data


if __name__ == '__main__':
    from cavass.ops import read_cavass_file, save_cavass_file
    import numpy as np

    image = read_cavass_file('/data1/dj/data/bca/cavass_data/images/N007PETCT.IM0')
    image = image[None].astype(np.float64)
    image = torch.from_numpy(image)
    data = {'image': image}

    gbt = MirrorTransform(keys=['image'], apply_probability=1, allowed_axes=[1], p_per_axes=1)
    data = gbt(data)

    image = data['image'][0].numpy()
    save_cavass_file('/data1/dj/tmp/image.IM0', image.astype(np.uint16),
                     reference_file='/data1/dj/data/bca/cavass_data/images/N007PETCT.IM0')
